
import os
import yaml
import itertools
import datetime
import logging

from collections import OrderedDict

from robob.util import time2sec
from robob.reporter import Reporter
from robob.metrics import Metrics
from robob.context import Context
from robob.stream import Stream, streamContext

def deepupdate(original, update):
	"""
	Recursively update a dict.
	Subdict's won't be overwritten but also updated.
	"""
	for key, value in original.items(): 
		if key not in update:
			update[key] = value
		elif isinstance(value, dict):
			deepupdate(value, update[key]) 
		elif isinstance(value, list):
			update[key] = value + update[key]
	return update

def ordered_load(stream, Loader=yaml.Loader, object_pairs_hook=OrderedDict):
	class OrderedLoader(Loader):
		pass
	def construct_mapping(loader, node):
		loader.flatten_mapping(node)
		return object_pairs_hook(loader.construct_pairs(node))
	OrderedLoader.add_constructor(
		yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG,
		construct_mapping)
	return yaml.load(stream, OrderedLoader)

class Specs(object):
	"""
	Specifications file with nested specifications resolution support
	"""

	def __init__(self, filename):
		"""
		Initialize a specifications object from the given filename
		"""

		self.filename = filename
		self.specs = OrderedDict()

	def getTestVariables(self):
		"""
		Return the variable names of the test-cases
		"""

		# Return keys of test-cases
		return list(self.specs['test-cases'].keys())

	def getMetricTitles(self):
		"""
		There is already some work done on the metrics class
		so we are going to re-use it
		"""

		# Create a metrics object
		metrics = Metrics()
		metrics.configure( self.context )

		# Return titles
		return metrics.titles()

	def createTestContexts(self):
		"""
		Create test contexts according to specs
		"""

		# Prepare product components
		values = []
		keys = []
		for k,v in self.specs['test-cases'].items():
			keys.append(k)
			values.append(v)

		# Generate test cases as product of combinations
		contexts = []
		for v in itertools.product(*values):

			# Fork context
			ctx = self.context.fork()

			# Update and collect
			test_keys = dict(list(zip( keys, v)))
			ctx.update( test_keys ) # Insert in global scope
			ctx.set( "curr", test_keys ) # And in curr scope
			contexts.append( ctx )

		return contexts

	def createStreams(self, testContext, testMetrics, iteration):
		"""
		Create a stream contexts using the specified test context as base
		"""
		ans = []


		# Create a stream object for every stream defined in specs
		for specs in self.specs['streams']:

			# Create and configure a stream
			stream = Stream( testContext, testMetrics, iteration )
			stream.configure( specs )

			# Append to list
			ans.append( stream )

		# Return streams
		return ans

	def createReporter(self):
		"""
		Create a reporter according to the specifications
		"""

		# Calculate filename
		filename = self.context['report.path'] + "/"
		filename += self.context['report.name']
		filename += "-%s" % self.context['report.timestamp']
		filename += ".csv"

		# Create reporter
		return Reporter( filename, self )

	def load(self):
		"""
		Load the specifications file
		"""
		logger = logging.getLogger("specs")

		# Prepare stack
		filestack = [ self.filename ]
		specsstack = [ ]

		# Start processing items on filestack
		while len(filestack):

			# Get filename and base dir to load
			fname = filestack.pop(0)
			logger.info("Loading %s" % fname)
			bdir = os.path.dirname( fname )
			if not bdir:
				bdir = "."

			# Load specs
			with open(fname, 'r') as f:
				buf = f.read()
				specs = ordered_load(buf, yaml.SafeLoader)

			# Check if there are other files to
			# load, and therefore add them on filestack
			if 'load' in specs:

				# Make sure it's list
				if type(specs['load']) in [str, str]:
					specs['load'] = [ specs['load'] ]

				# Iterate over specs
				for f in specs['load']:
					if f[0] == "/":
						filestack.append(f)
					else:
						filestack.append( "%s/%s" % (bdir, f) )

				# Remove 'load'
				del specs['load']

			# Keep specs
			specsstack.append( specs )

		# Merge specs in reverse order so that loaded
		# files have lower priority than the ones that loaded them
		for specs in reversed( specsstack ):
			self.specs = deepupdate( self.specs, specs )

		# Open a global context & import global variables
		self.context = Context()
		if 'globals' in self.specs:
			self.context.update( self.specs['globals'] )

		# Apply test specs
		if 'test' in self.specs:
			self.context.set( 'test', self.specs['test'] )

		# Import environments
		if 'environments' in self.specs:
			for k,v in self.specs['environments'].items():
				self.context.set("env", self.specs['environments'])

		# Import metrics
		if 'metrics' in self.specs:
			self.context.set("metric", self.specs['metrics'])

		# Import nodes
		if 'nodes' in self.specs:
			self.context.set( 'node', self.specs['nodes'] )

		# Import parsers
		if 'parsers' in self.specs:
			self.context.set( 'parser', self.specs['parsers'] )

		# Import apps
		if 'apps' in self.specs:
			self.context.set( 'app', self.specs['apps'] )

		# Import streamlets
		if 'streamlets' in self.specs:
			self.context.set( 'streamlet', self.specs['streamlets'] )

		# Import notes
		if 'notes' in self.specs:
			self.context.set( 'notes', self.specs['notes'] )

		# Import report
		if 'report' in self.specs:
			self.context.set("report", self.specs['report'])

		# Initialize report defaults
		if not 'report.name' in self.specs:
			name = "test"
			if 'name' in self.specs:
				name = self.specs['name']
			self.context.set('report.name', name)
		if not 'report.path' in self.specs:
			baseDir = "."
			if os.path.isdir("./reports"):
				baseDir = "./reports"
			self.context.set('report.path', baseDir)

		# Define report timestamp
		d = datetime.datetime.now()
		self.context.set( 'report.timestamp', d.strftime("%Y%m%d%H%M%S") )




import random
import string
from robob.pipe import PipeBase

class Pipe(PipeBase):
	"""
	Implementation of the file generator pipe item
	"""

	def configure(self, config):
		"""
		Configure cmdline
		"""

		# Prepare file generator
		self.name = config['name']
		self.path = config['path']
		self.contents = config['contents']

	def pipe_cmdline(self):
		"""
		Return script components as cmdline
		"""

		# A unique contents separator
		eof_indicator = "CONTENTS_"
		eof_indicator += ''.join(random.choice(string.ascii_uppercase + string.digits) for _ in range(16))

		# Prepare script
		genscript  = "WFILE=\"%s\"\n" % self.path
		genscript += "cat <<'%s' > $WFILE\n" % eof_indicator
		genscript += self.contents
		genscript += "\n%s\n" % eof_indicator

		# Return script that generates the given file
		return [ "eval", genscript ]




import random
import string
from robob.pipe import PipeBase

class Pipe(PipeBase):
	"""
	Implementation of the file deleting pipe item
	"""

	def configure(self, filename):
		"""
		Configure cmdline
		"""

		# The filename to delete
		self.name = filename

	def pipe_cmdline(self):
		"""
		Return script components as cmdline
		"""

		# Return script that deletes the given file if it exists
		delscript = "[ -f \"%s\" ] && rm \"%s\"" % (self.name, self.name)
		return [ "eval", delscript ]




import logging
import re
import copy

from collections import OrderedDict

#: Macro regex
RE_MACRO = re.compile(r'\$\{(.+?)\}')

#: Macro variable regex
RE_MACRO_VAR = re.compile(r'[a-z][a-z0-9\._]*', flags=re.IGNORECASE)

#: Context logger
logger = logging.getLogger('context')

class Context(OrderedDict):
	"""
	An environment variable and context
	"""

	def fork(self):
		"""
		Deep copy current context into another context
		"""
		return Context( copy.deepcopy(self) )

	def set(self, name, value, include_flag=True):
		"""
		Set the specified value in the context and optionally
		also update a flat representation of the value.
		"""

		# Set value
		self[name] = value

		# Update flat representation
		if include_flag:
	
			if isinstance(value, dict):
				for k,v in value.items():
					self.set( "%s.%s" % (name,k), v )

			elif isinstance(value, list):
				i = 0
				for v in value:
					if 'name' in v:
						self.set( "%s.%s" % (name,v['name']), v )
					else:
						self.set( "%s.%i" % (name,i), v )
					i += 1

	def merge(self, name, value, include_flag=True):
		"""
		Merge the specified value in the context and optionally
		also update a flat representation of the value.
		"""

		# Set value
		if name in self:
			if isinstance(value, dict):

				# Merge list
				i = len(self[name])
				self[name] += value

				# Update flat representation
				if include_flag:
					for v in value:
						if 'name' in v:
							self.set( "%s.%s" % (name,v['name']), v )
						else:
							self.set( "%s.%i" % (name,i), v )
						i += 1

				# Don't continue
				return

			elif isinstance(value, list):

				# Merge dictionary
				self[name].update( value )

				# Update flat representation
				if include_flag:
					for k,v in value.items():
						self.set( "%s.%s" % (name,k), v )

				# Don't continue
				return

		# Update value
		self[name] = value

	def render(self):
		"""
		Replace all macros in this context and return a
		dictionary with all values
		"""
		self._unreplaced = set()
		dictionary = copy.deepcopy(self)
		replaced = True

		# Keep replacing until there are no other macros to replace
		while replaced:
			# Replace and check if this was indeed replaced
			(dictionary, replaced) = self.replaceMacros( dictionary )

		# Log unreplaced items
		for m in self._unreplaced:
			logger.warn("Unknown macro '${%s}' encountered in specifications!" % m)

		# Return a new context with the new dictionary
		return Context( dictionary )

	def evaluate(self, expr):
		"""
		Evaluate a macro expression
		"""

		# Reset properties
		self._evalMissing = False

		# Extract default value
		defaultValue = None
		if '|' in expr:
			(expr, defaultValue) = expr.split("|", 1)

		# Check if this is an expression
		if ('*' in expr) or ('+' in expr) or ('/' in expr) or \
			('-' in expr) or ('^' in expr) or ('(' in expr) or (')' in expr):

			# Replace helper
			def replace(m):
				key = m.group(0)

				# Ignore some functions
				if key in ['str', 'int', 'float']:
					return m.group(0)

				# Replace values
				if not key in self:
					self._evalMissing = True
					return ""
				else:

					# Check if we should wrap value in quotes
					value = str(self[key])
					if value.isdigit() or ((value.count(".") == 1) and value.replace(".","").isdigit()):
						return value
					else:
						return '"%s"' % value.replace('"', '\\"')

			# Expand variables
			eval_expr = RE_MACRO_VAR.sub( replace, expr )

			# Check if we had missing properties
			if self._evalMissing:
				return defaultValue

			# Otherwise evaluate expression
			try:
				return eval(eval_expr)
			except Exception as e:
				logger.warn("Error evaluating macro expression '%s': %s" % (expr, str(e)))
				return defaultValue

		else:

			# It's a plain value, return calculation
			if expr in self:
				return self[expr]
			else:
				return defaultValue

	def replaceMacros(self, where, _firstCall=True):
		"""
		Replace all macros in context to the given string
		"""
		unused = False

		# Check if a value was indeed updated
		if _firstCall:
			self._did_replace = False

		# Replace helper
		def replace(m):
			# Try to evaluate expression
			key = m.group(1)
			value = self.evaluate( key )
			if value is None:
				# Return un-replaced
				self._unreplaced.add(key)
				return m.group(0)
			else:
				# Return replaced & mark action
				self._did_replace = True
				# Remove from unreplaced
				if key in self._unreplaced:
					self._unreplaced.remove(key)
				return str(value)

		# Replace all macros in dict
		if isinstance(where, dict):
			ans = {}
			for k,v in where.items():
				(ans[k], unused) = self.replaceMacros(v, False)
			return (ans, self._did_replace)

		# Replace all macros in list
		elif isinstance(where, list):
			ans = []
			for v in where:
				ans.append( self.replaceMacros(v, False)[0] )
			return (ans, self._did_replace)

		# Replace all macros in string
		elif type(where) in [str, str]:
			return (RE_MACRO.sub( replace, where ), self._did_replace)

		# Pass through everything else
		else:
			return (where, self._did_replace)





from robob.pipe import PipeListener

class LogPipe(PipeListener):
	"""
	This class is used internally to log all stdout lines into a file
	"""

	def __init__(self, filename):
		"""
		"""
		PipeListener.__init__(self)
		self.fd = open(filename, "w")

	def got_stdout(self, line):
		"""
		Match the specified line against our configuration
		"""
		self.fd.write("%s\n" % line)

	def got_eof(self):
		"""
		Process end of stream
		"""
		self.fd.close()




def time2sec(timestr):
	"""
	Conver time specs to seconds
	"""

	# Skip numerics
	if type(timestr) in [int, float]:
		return timestr

	# Process strings
	if timestr[-1] == "s":
		return float(timestr[0:-1])
	elif timestr[-1] == "m":
		return float(timestr[0:-1]) * 60
	elif timestr[-1] == "h":
		return float(timestr[0:-1]) * 60 * 60
	else:
		return float(timestr)




from robob.pipe import PipeListener
from robob.component import ComponentBase

class ParserBase(ComponentBase, PipeListener):
	"""
	Base class for implementing stream output parser
	"""

	def __init__(self, ctx, metrics):
		"""
		Initialize parser base class
		"""
		ComponentBase.__init__(self, ctx)
		PipeListener.__init__(self)

		# Initialize metrics
		self._metrics = metrics
		self._alias = {}
		self._filter = None

	def got_stdout(self, line):
		"""
		[Public] Process an stdout line
		"""
		pass

	def got_stderr(self, line):
		"""
		[Public] UProcess an stderr line
		"""
		pass

	def reset(self):
		"""
		[Public] Reset the parser for a new stream
		"""
		self._metrics.reset()

	def set_alias(self, aliases):
		"""
		[Public] Update alias mapping
		"""
		self._alias = aliases

	def set_filter(self, metrics):
		"""
		[Public] Update metrics filter
		"""
		self._filter = metrics

	def update(self, metric, value):
		"""
		[Private] Update the value of the specified metric
		"""

		# Skip metrics in filter
		if not self._filter is None and not metric in self._filter:
			return

		# Replace alias if found
		if metric in self._alias:
			metric = self._alias[metric]

		# Forward to metrics
		self._metrics.update( metric, value )




import sys
import os
import time
import signal
import logging
import robob.logger

from robob.util import time2sec
from robob.specs import Specs
from robob.driver import TestDriver

def help(verbose=False):
	"""
	Show help screen
	"""
	print("RoBOB - Simplify collection of measurements over repetitive tasks")
	print("Read more: https://github.com/wavesoft/robob/wiki")
	print("")
	print("Usage: robob <path-to-benchmark.yaml>")
	print("")
	sys.exit(1)

def main():

	# Get a logger
	logger = logging.getLogger("robob")

	# Show help screen if missing arguments
	if len(sys.argv) < 2:
		help()
		return 2

	# Check for help
	if sys.argv[1] in ['-h', '--help']:
		help(True)
		return 0

	# Validate file
	specsfile = sys.argv[1]
	if not os.path.isfile(specsfile):
		logger.error("The specified file '%s' was not found!" % specsfile)
		help()
		return 1

	# Load specs
	specs = Specs( specsfile )
	specs.load()

	# Create test contexts
	tests = specs.createTestContexts()

	# Create reporter
	reporter = specs.createReporter()
	reporter.start()

	# Gracefully shutdown
	def cleanup(signal, frame):

		# Interrupt reporter
		logger.warn("Received break signal from the user")	

		# Interrupt driver and reporter
		driver.interrupt()
		reporter.interrupt( driver.summarize() )
		reporter.close()

		# Exit with error
		sys.exit(1)

	# Trap shutdown signal
	signal.signal(signal.SIGINT, cleanup)

	# Create stream context for every test
	driver = None
	test_id = 0
	for test in tests:

		# Create a test driver
		driver = TestDriver( specs, test )

		# Get some values from test specs
		iterations = int(test.get("test.iterations", 1))
		cooldown = time2sec(test.get("test.cooldown", 0))

		# Start reporting the test
		reporter.test_start( test )

		# Run multiple iterations of the test
		for i in range( 0, iterations ):

			# Calculate progress
			p_total = (len(tests) * iterations)
			p_curr = test_id * iterations + i

			# Start log
			logger.info("Running %i/%i (test: %i/%i, iteration: %i/%i, values: {%s})" % (p_curr+1, p_total, test_id+1, len(tests), i+1, iterations, \
				", ".join([ "%s=\"%s\"" % (k, str(v)) for k,v in test['curr'].iteritems() ]) ))
			reporter.iteration_start( i+1 )

			# Run driver
			driver.run(i)
			reporter.iteration_end( driver.lastResults, driver.lastStatus, driver.lastComment )

			# Apply cooldown
			if cooldown:
				logger.info("Waiting for %s sec before next test" % test.get("test.cooldown", "0"))
				time.sleep(cooldown)

		# Summarize iterations and finalize test
		reporter.test_end( driver.summarize() )

		# Increment test ID
		test_id += 1

	# Finalize reporter
	reporter.finalize()
	reporter.close()
	return 0




import shlex
from robob.pipe import PipeBase

class Pipe(PipeBase):
	"""
	Implementation of the script pipe item
	"""

	def configure(self, config):
		"""
		Configure cmdline
		"""

		# Prepare cmdline
		self.cmdline = ["eval", config['script']]

	def pipe_cmdline(self):
		"""
		Return script components as cmdline
		"""
		return self.cmdline




import logging

(BLACK, RED, GREEN, YELLOW, BLUE, MAGENTA, CYAN, WHITE) = list(range(8))

#These are the sequences need to get colored ouput
RESET_SEQ = "\033[0m"
COLOR_SEQ = "\033[1;%dm"
BOLD_SEQ = "\033[1m"

#: Color lookup table for formatter
COLORS = {
	'WARNING': YELLOW,
	'INFO': GREEN,
	'DEBUG': BLUE,
	'CRITICAL': YELLOW,
	'ERROR': RED
}

#: Aliases for constant-sized names
ALIASES = {
	'WARNING' : 'WARN',
	'INFO'	  : 'INFO',
	'DEBUG'	  : 'DEBG',
	'CRITICAL': 'CRIT',
	'ERROR'	  : 'FAIL'
}

def formatter_message(message, use_color=True):
	"""
	Message formatter that expands $RESET and $BOLD macros
	"""
	if use_color:
		message = message.replace("$RESET", RESET_SEQ).replace("$BOLD", BOLD_SEQ)
	else:
		message = message.replace("$RESET", "").replace("$BOLD", "")
	return message

class ColoredFormatter(logging.Formatter):
	"""
	Colored formatter
	"""

	def __init__(self, msg, use_color = True):
		logging.Formatter.__init__(self, msg)
		self.use_color = use_color

	def format(self, record):
		"""
		Format the specified log line
		"""

		levelname = record.levelname
		if self.use_color and levelname in COLORS:

			# Add color to level name
			levelname_color = COLOR_SEQ % (30 + COLORS[levelname]) + ALIASES[levelname] + RESET_SEQ
			record.levelname = levelname_color

			# Make name
			record.name = COLOR_SEQ % (30 + WHITE) + record.name + RESET_SEQ

		return logging.Formatter.format(self, record)

# Custom logger class with multiple destinations
class RobobLogger(logging.Logger):

	FORMAT = "[$BOLD%(levelname)s$RESET][%(name)s] %(message)s"
	COLOR_FORMAT = formatter_message(FORMAT, True)

	def __init__(self, name):
		logging.Logger.__init__(self, name, logging.INFO)                

		# Use colored formatter
		color_formatter = ColoredFormatter(self.COLOR_FORMAT)

		# Add console target
		console = logging.StreamHandler()
		console.setFormatter(color_formatter)
		self.addHandler(console)

# Set robob logger
logging.setLoggerClass(RobobLogger)




import errno
import time
import signal
import select
import logging

import pty, os, fcntl, termios

from threading import Thread
from subprocess import Popen, PIPE
from robob.metrics import Metrics, summarize

class PtyProcess:

	def __init__(self, cmdline, **kwargs):
		"""
		Initialize a pty process control
		"""

		self.cmdline = cmdline
		self.kwargs = kwargs
		self.returncode = None
		self.fd = None

		# Fork into a new pty
		self.pid, self.fd = pty.fork()
		if self.pid == 0:
			self._init_child()
		else:
			self._init_host()

	def _init_child(self):
		"""
		Child process
		"""
		os.execv( self.cmdline[0], self.cmdline )

	def _init_host(self):
		"""
		Host process
		"""

		# Disable echo in the tty
		new = termios.tcgetattr(self.fd)
		new[3] = new[3] & ~termios.ECHO
		termios.tcsetattr(self.fd, termios.TCSADRAIN, new)

	def close(self):
		"""
		Close FDs
		"""

		if self.fd:
			if self.returncode != None:
				os.close(self.fd)
			self.fd = None

	def send_signal(self, sig):
		"""
		Send signal to process
		"""
		try:
			os.kill(self.pid, sig)
		except (OSError, IOError) as e:
			return

	def terminate(self):
		"""
		Send kill signal to process
		"""
		try:
			os.kill(self.pid, signal.SIGTERM)
		except (OSError, IOError) as e:
			return

	def poll(self):
		"""
		Check if thread has exited
		"""

		# Get PID
		if self.returncode is None:
			try:
				pid, sts = os.waitpid(self.pid, os.WNOHANG)
				if pid == self.pid:
					self._handle_exitstatus(sts)
			except os.error as e:
				self.returncode = None

		return self.returncode

	def wait(self):
		"""
		Wait for thread to exit and return exit code
		"""

		# Get PID
		while self.returncode is None:
			try:
				pid, sts = os.waitpid(self.pid, 0)
				if pid == self.pid:
					self._handle_exitstatus(sts)
			except (OSError, IOError) as e:
				if e.errno == errno.EINTR:
					continue
			except os.error as e:
				self.returncode = None

		return self.returncode

	def _handle_exitstatus(self, sts, _WIFSIGNALED=os.WIFSIGNALED,
			_WTERMSIG=os.WTERMSIG, _WIFEXITED=os.WIFEXITED,
			_WEXITSTATUS=os.WEXITSTATUS):
		# This method is called (indirectly) by __del__, so it cannot
		# refer to anything outside of its local scope.
		if _WIFSIGNALED(sts):
			self.returncode = -_WTERMSIG(sts)
		elif _WIFEXITED(sts):
			self.returncode = _WEXITSTATUS(sts)
		else:
			# Should never happen
			raise RuntimeError("Unknown child exit status!")


class TestStreamThread(Thread):
	"""
	A thread that runs a test
	"""

	def __init__(self, stream, *args, **kwargs):
		"""
		Initialize a test launcher
		"""
		Thread.__init__(self, *args, **kwargs)

		self.proc = None
		self.stream = stream
		self.returncode = 0
		self.interrupted = False
		self.interruptReason = ""
		self.logger = logging.getLogger("stream.%s" % self.stream.name)
		self.lastactivity = time.time()

	def interrupt(self, reason="User interrupt", timeout=5):
		"""
		Interrupt the subprocess
		"""

		# Skip if already interrupted
		if self.interrupted:
			return

		# Mark as interrupted
		self.logger.critical("Interrupting stream")
		self.interrupted = True
		self.interruptReason = reason

		# First gracefully kill the sub-process
		if self.proc:

			# First send sigint
			self.proc.send_signal( signal.SIGINT )

			# Wait 5 seconds to 
			t_timeout = time.time() + timeout
			while True:

				# Check if this was enough
				if self.proc.poll() != None:
					self.logger.warn("Stream interrupted")
					break

				# Check for timeouts
				if time.time() > t_timeout:

					# Terminate
					self.proc.terminate()
					self.logger.warn("Stream terminated")
					break

			# Reap process
			self.proc.close()
			self.proc = None

			# Set return code to -1
			self.returncode = -1

	def run(self):
		"""
		Main run function
		"""
		pipe = self.stream.pipe

		# Apply delay
		if self.stream.delay:
			self.logger.info("Delaying for %i seconds" % self.stream.delay)
			q_time = time.time() + self.stream.delay
			while time.time() < q_time:
				time.sleep(0.1)
				if self.interrupted:
					self.logger.warn("Interrupted while in delay")
					return

		# Prepare expiration time
		t_expire = None
		if self.stream.timeout:
			t_expire = time.time() + self.stream.timeout

		# Open process
		self.logger.debug("Process starting %r" % pipe.pipe_cmdline())
		self.proc = proc = PtyProcess( pipe.pipe_cmdline() )

		# Get a sequence of optional expect entries
		expect_out = pipe.pipe_expect_stdout()
		expect_err = pipe.pipe_expect_stderr()
		self.has_expect = True

		# Send stdin if there are no expect entries
		if (len(expect_out) == 0) and (len(expect_err) == 0):
			self.logger.debug("Sending STDIN payload")
			try:
				os.write( proc.fd, pipe.pipe_stdin() )
				os.write( proc.fd, "\x04") # End-of-transmission
			except (OSError, IOError) as e:
				self.logger.warn("%s occured: %s" % (e.__class__.__name__, str(e)))
				self.interrupt("%s: %s" % (e.__class__.__name__, str(e)))
				return
			self.has_expect = False

		# Flag that keeps line polling activ
		active = True

		# Helper function to handle lines
		def handle_line(read):

			# Consider this an "activity" action
			self.lastactivity = time.time()

			# Skip empty lines
			if not read.strip():
				return

			# Ignore '\r'
			read = read.replace("\r", "")

			# Guard against exceptions
			self.logger.debug("STDIN: %s" % read)
			try:

				# First apply over expect
				handled = False
				for i in range(0, len(expect_out)):
					if expect_out[i].matches( read ):
						self.logger.debug("Expect matched as /%s/ on stdin" % str(expect_out[i]))
						os.write( proc.fd, expect_out[i].render() )
						del expect_out[i]
						handled = True
						break

				# If not handled, pass to pipe
				if not handled:
					pipe.pipe_stdout( read )

			except Exception as e:

				import traceback
				traceback.print_exc()

				# In case something went wrong while processing the streams,
				# kill the process
				self.logger.error("%s occured: %s" % (e.__class__.__name__, str(e)) )
				active = False

				# Interrupt
				self.interrupt("%s: %s" % (e.__class__.__name__, str(e)))
				return

			# If we have processed all expect entries, send STDIN
			if self.has_expect and (len(expect_out) == 0) and (len(expect_err) == 0):
				self.logger.debug("No more expects left, sending STDIN payload")
				os.write( proc.fd, pipe.pipe_stdin() )
				os.write( proc.fd, "\x04") # End-of-transmission
				self.has_expect = False

		# Read stdout/err
		self.logger.debug("Processing output")
		data = ""
		data_flush_t = 0
		while active:
			ts = time.time()

			# If interrupted, quit
			if self.interrupted:
				pipe.pipe_close()
				return

			# Forward incomplete data as incomplete line
			if data and (ts > data_flush_t):
				handle_line(data)
				data = ""

			# Wait for data within 100ms
			if proc.fd in select.select([proc.fd], [], [], 0.1)[0]:

				# If interrupted, quit
				if self.interrupted:
					pipe.pipe_close()
					return

				# Read data
				buf = os.read( proc.fd, 4096)
				if buf:

					# Stack buffers
					data += buf

					# Replace windows newlines & process lines
					data = data.replace("\r\n", "\n")
					while "\n" in data:
						(line, data) = data.split("\n",1)
						handle_line(line)

					# Note when we received the data
					# in order to forward them as-is as
					# incomplete line data after a timeout
					data_flush_t = ts + 0.1

			# Check for process exit
			if proc.poll() != None:
				self.logger.debug("Process exited with code %i" % proc.returncode)
				break

			# Check if timeout expired
			if t_expire and (ts > t_expire):
				self.logger.critical("Timeout of %s seconds expired" % self.stream.timeout)
				self.interrupt("Timeout after %s sec" % self.stream.timeout)
				pipe.pipe_close()
				return

			# Check if idle timeout expired
			if self.stream.idletimeout and (ts > (self.lastactivity + self.stream.idletimeout)):
				self.logger.critical("Timeout after %s seconds of inactivity" % self.stream.idletimeout)
				self.interrupt("Timeout after %s seconds of inactivity" % self.stream.idletimeout)
				return

		# If not interrupted, clean shutdown
		if not self.interrupted and self.proc:
			pipe.pipe_close()
			proc.close()
			self.returncode = proc.returncode
			self.proc = None

		# Cleanup
		if self.returncode != 0:
			self.logger.warn("Stream closed with error code=%i" % proc.returncode)
		else:
			self.logger.info("Stream closed successfuly")

class TestDriver:
	"""
	This class is responsible for starting the streams, monitoring
	their output, feeding them to the appropriate stream readers 
	and collecting the test results.
	"""

	def __init__(self, specs, test):
		"""
		Initialize a new test driver objet
		"""

		self.specs = specs
		self.test = test
		self.logger = logging.getLogger("driver")

		self.metrics = Metrics()
		self.metrics.configure( test )

		self.results = []
		self.lastResults = None
		self.lastStatus = ""
		self.lastComment = ""

	def run(self, iteration):
		"""
		Start the tests on the test driver
		"""

		# Create all stream specifications on this context
		streams = self.specs.createStreams( self.test, self.metrics, iteration )

		# Reset metrics
		self.metrics.reset()
		self.lastStatus = "Completed"
		self.lastComment = ""

		# Launch them
		self.threads = []
		self.logger.debug("Starting %i streams" % len(streams))
		for s in streams:

			# Skip inactive streams
			if not s.active:
				self.logger.info("Not starting inactive stream '%s'" % s.name)
				continue

			# Create a test stream thread
			t = TestStreamThread( s )

			# Start and keep
			t.start()
			self.threads.append(t)

		# Wait for all threads to complete
		self.logger.debug("Waiting for stream threads to exit")
		hasAlive = True
		hasInterrupt = False
		while hasAlive:

			# Check status of all items
			hasAlive = False
			for t in self.threads:
				if t.interrupted:
					hasInterrupt = True
					break
				if t.is_alive():
					hasAlive = True
					break

			# If something interrupted, interrupt everything
			if hasInterrupt:
				self.logger.warn("Thread '%s' interrupted, so collapsing this test-case" % t.stream.name)
				self.interrupt(t.interruptReason)
				break

			# Otherwise wait for a sec before retry
			if hasAlive:
				time.sleep(0.1)

		# Update exit code status
		for t in self.threads:
			if t.returncode != 0:
				if self.lastComment:
					self.lastComment += "; "
				self.lastComment += "%s returned=%i" % (t.stream.name, t.returncode)

				# Switch to error only if not completed (it might be timeout for example)
				if self.lastStatus == "Completed":
					self.lastStatus = "Error"

		# Reap threads
		for t in self.threads:
			t.join()

		# Reset thread list
		self.threads = []

		# Collect results
		self.logger.debug("Threads exited, collecting results")
		self.lastResults = self.metrics.results()
		self.results.append( self.lastResults )

	def summarize(self):
		"""
		Summarize results
		"""

		# Summarize results
		return summarize( self.results )

	def interrupt(self, reason="Interrupted"):
		"""
		Interrupt current run
		"""

		# Collect results
		self.logger.warn("Driver interrupted (%s), stopping all threads" % reason)
		self.lastStatus = reason
		self.lastResults = self.metrics.results()
		self.results.append( self.lastResults )

		# Interrupt and join all threads
		for t in self.threads:
			self.logger.debug("Interrupting stream thread %s" % t.stream.name)
			t.interrupt()
			self.logger.debug("Joining stream thread %s" % t.stream.name)
			t.join()




import os
import re
import logging
import random
import string

from robob.util import time2sec
from robob.factories import pipeFactory, parserFactory
from robob.metrics import Metrics
from robob.logpipe import LogPipe
from robob.pipe.bashwrap import Pipe as BashWrapPipe
from robob.pipe.app import Pipe as AppPipe
from robob.pipe.filegen import Pipe as FileGenPipe
from robob.pipe.filedel import Pipe as FileDelPipe

def streamContext( context, specs ):
	"""
	Update given context for use by the stream with the given specs
	"""

	# Fork context
	context = context.fork()

	# Get node
	node = specs['node']
	if not "node.%s" % node in context:
		raise AssertionError("Node '%s' was not defined in the specs" % node)
	node = context["node.%s" % node]

	# Get app
	app = specs['app']
	if not "app.%s" % app in context:
		raise AssertionError("App '%s' was not defined in the specs" % app)
	app = context["app.%s" % app]

	# Get optional app env
	env = None
	if 'env' in app:
		env = app['env']
		if not "env.%s" % env in context:
			raise AssertionError("env '%s' was not defined in the specs" % env)
		env = context["env.%s" % env]

	########################################
	# Initialize context variables
	########################################

	# Define context of current node, parser, app
	context.set( "stream", specs )
	context.set( "node", node )
	context.set( "app", app )
	if env:
		context.set( "env", env )
		context.update( env )

	# Update custom variable definitions
	if 'define' in node:
		context.update( node['define'] )
	if 'define' in app:
		context.update( app['define'] )
	if 'define' in specs:
		context.update( specs['define'] )

	# Update app file paths if temporary
	if 'app.files' in context:
		for f in context['app.files']:

			# Create filename of temporary file
			if not 'path' in f:

				# Calcuate suffix
				suffix = ".tmp"
				if 'suffix' in f:
					suffix = f['suffix']

				# Calculate temporary path
				tmp_path = "/tmp/robob.%s-%s%s" % (
						f['name'],
						''.join(random.choice(string.ascii_lowercase + string.digits) for _ in range(24)),
						suffix
					)

				# Update context
				f['path'] = tmp_path
				f['temp'] = True
				context.set("app.files.%s.path" % f['name'], tmp_path)
				context.set('app.files.%s.temp' % f['name'], True)

			else:

				# Mark as non-temporary
				f['temp'] = False
				context.set('app.files.%s.temp' % f['name'], False)

	# Render and return context
	return context.render()

RE_SANITIZE = re.compile(r"[^A-Za-z0-9]+")

def sanitize_fname(v):
	"""
	Sanitize fileame
	"""
	return RE_SANITIZE.sub("-", str(v))

class Stream(object):
	"""
	A stram on which tests can run
	"""

	# Last stream ID
	LAST_STREAM_ID = 0

	def __init__(self, context, metrics, iteration):
		"""
		Initialize a new stream
		"""

		self.delay = 0
		self.name = "stream_%i" % Stream.LAST_STREAM_ID
		self.pipe = None
		self.bashPipe = None
		self.appPipe = None
		self.accessPipe = None
		self.metrics = metrics
		self.context = context
		self.timeout = None
		self.idletimeout = None
		self.active = True
		self.iteration = iteration

		# Open logger
		self.logger = logging.getLogger("stream.%s" % self.name)

		# Update last stream ID
		Stream.LAST_STREAM_ID += 1

	def openLogPipe(self):
		"""
		Open a pipe to the filename that will hold the output of the application
		"""

		# If missing, return none
		if not 'report.keep_output' in self.context:
			return None

		# Calculate filename
		testval = "+".join([ "%s-%s" % (k, sanitize_fname(v)) for k,v in self.context['curr'].items() ])
		filename = "out-%s-%s-%i" % (self.name, testval, self.iteration+1)
		filename += ".log"

		# Calculate directory name
		basedir = self.context['report.keep_output'] + "/"
		basedir += self.context['report.name']
		basedir += "-%s" % self.context['report.timestamp']

		# Make directory
		if not os.path.exists(basedir):
			os.mkdir(basedir)
		elif not os.path.isdir(basedir):
			raise AssertionError("Directory %s is not directory!" % basedir)

		# Log
		filename = "%s/%s" % (basedir, filename)
		self.logger.info("Logging STDOUT to %s" % filename)

		# Create and return a new logpipe
		return LogPipe(filename)

	def configure(self, specs):
		"""
		Configure stream from the specified specs context
		"""

		# Get simple properties
		self.delay = 0
		if 'delay' in specs:
			self.delay = time2sec(specs['delay'])
		if 'timeout' in specs:
			self.timeout = time2sec(specs['timeout'])
		if 'idle' in specs:
			self.idletimeout = time2sec(specs['idle'])
		if 'name' in specs:
			self.name = specs['name']
			self.logger = logging.getLogger("stream.%s" % self.name)

		# Initialize context
		self.context = streamContext( self.context, specs )

		# Check if stream is active
		if 'stream.active' in self.context:
			self.active = self.context['stream.active']
			if type(self.active) in [str, str]:
				self.active = self.active.lower() in [ "1", "yes", "true", "on" ]

		# If not active exit
		if not self.active:
			return

		########################################
		# Initialize pipes
		########################################

		# Factory app pipe
		self.appPipe = AppPipe( self.context )
		self.appPipe.configure( self.context['app'] )

		# Factory bash multiplexer/wrapper
		self.bashPipe = BashWrapPipe( self.context )
		self.bashPipe.plug( self.appPipe )

		# Factory log pipe to capture output
		out = self.openLogPipe()
		if out:
			self.appPipe.listen( out )

		########################################
		# Initialize file generators
		########################################

		# Instantiate file generators
		if 'app.files' in self.context:
			for filegen in self.context['app.files']:

				# Instantiate and configure filegen pipe
				fpipe = FileGenPipe( self.context )
				fpipe.configure( filegen )

				# Add precondition to bash pipe
				self.bashPipe.plug_pre( fpipe )

				# If this is temporary, add cleanup pipe
				if filegen['temp']:

					# Instantiate and configure file delete pipe
					dpipe = FileDelPipe( self.context )
					dpipe.configure( fpipe.path )

					# Add post-condition to bash pipe
					self.bashPipe.plug_post( dpipe )

		########################################
		# Initialize parsers
		########################################

		# Instantiate app parsers
		parser_names = []
		if 'app.parser' in self.context:
			parser_names.append( self.context['app.parser'] )
		elif 'app.parsers' in self.context:
			parser_names += self.context['app.parsers']
		else:
			raise AssertionError("It's required to define at least one parser on app '%s'" % self.context['app.name'])

		# Instantiate parsers
		for n in parser_names:
			if not "parser.%s" % n in self.context:
				raise AssertionError("Parser '%s' was not defined in the specs" % n)

			# Factory parser
			parser = parserFactory(self.context["parser.%s" % n], self.context, self.metrics )

			# Apply alias mapping & filter if exists
			if 'stream.alias' in self.context:
				self.logger.debug("Adding alias mapping to %s: %r" % (n, self.context['stream.alias']))
				parser.set_alias( self.context['stream.alias'] )
			if 'stream.filter' in self.context:
				self.logger.debug("Adding metrics filter to %s: %r" % (n, self.context['stream.filter']))
				parser.set_filter( self.context['stream.filter'] )

			# Listen for app output
			self.logger.debug("Adding parser %s to app listeners" % n)
			self.appPipe.listen( parser )

		# Instantiate streamlets
		if 'streamlets' in specs:
			for slt in specs['streamlets']:

				# Expand string shorthand
				if type(slt) in [str, str]:
					slt = { "streamlet": slt }

				# Get name
				if not 'streamlet' in slt:
					raise AssertionError("Missing 'streamlet' keyword in specs of stream '%s'" % self.name)
				n = slt['streamlet']

				# Get streamlet
				if not "streamlet.%s" % n in self.context:
					raise AssertionError("Streamlet '%s' was not defined in specs" % n)
				streamlet = self.context["streamlet.%s" % n]

				# Create a streamlet context & merge definitions
				streamlet_context = self.context.fork()
				streamlet_context.set( "streamlet", streamlet )
				streamlet_context.set( "streamlet", slt )

				# Render context
				streamlet_context = streamlet_context.render()

				# Instantiate streamlet pipe
				if not 'class' in streamlet:
					streamlet['class'] = "robob.pipe.script"
				pipe = pipeFactory( streamlet, streamlet_context )

				# Plug it on bash pipe
				self.logger.debug("Adding streamlet %s" % n)
				self.bashPipe.plug( pipe )

				# Get parser(s)
				parser_names = []
				if 'streamlet.parser' in streamlet_context:
					parser_names.append( streamlet_context['streamlet.parser'] )
				elif 'streamlet.parsers' in streamlet_context:
					parser_names += streamlet_context['streamlet.parsers']

				# Instantiate parsers
				for n in parser_names:
					if not "parser.%s" % n in streamlet_context:
						raise AssertionError("Parser '%s' was not defined in the specs" % n)

					# Factory parser
					parser = parserFactory(streamlet_context["parser.%s" % n], streamlet_context, self.metrics )

					# Apply stream alias mapping & filter if exists
					if 'stream.alias' in streamlet_context:
						self.logger.debug("Adding alias mapping to %s: %r" % (n, streamlet_context['stream.alias']))
						parser.set_alias( streamlet_context['stream.alias'] )
					if 'stream.filter' in streamlet_context:
						self.logger.debug("Adding metrics filter to %s: %r" % (n, streamlet_context['stream.filter']))
						parser.set_filter( streamlet_context['stream.filter'] )

					# Apply streamlet alias mapping & filter if exists
					if 'streamlet.alias' in streamlet_context:
						self.logger.debug("Adding streamlet alias mapping to %s: %r" % (n, streamlet_context['streamlet.alias']))
						parser.set_alias( streamlet_context['streamlet.alias'] )
					if 'streamlet.filter' in streamlet_context:
						self.logger.debug("Adding streamlet metrics filter to %s: %r" % (n, streamlet_context['streamlet.filter']))
						parser.set_filter( streamlet_context['streamlet.filter'] )

					# Factory parser & listen for app output
					self.logger.debug("Adding parser %s to streamlet listeners" % n)
					pipe.listen( parser )

		########################################
		# Initialize host accessor
		########################################

		# Get node configuration
		node = self.context['node']
		if not 'access' in node:
			raise AssertionError("Required at least one access component on node specs")

		# Create and chain accessor component
		for a in node['access']:

			# Merge node components into the accessor configuration
			specs = dict(node)
			specs.update( a )

			# Create accessor pipe
			pipe = pipeFactory( specs, self.context )

			# Chain them all the way to the bash pipe
			if self.accessPipe is None:
				self.accessPipe = pipe
				self.accessPipe.plug( self.bashPipe )
			else:
				pipe.plug( self.accessPipe )
				self.accessPipe = pipe

		# That's now our master pipe and we are ready to go!
		self.pipe = self.accessPipe






import importlib

def _class_by_name( module, className ):
	"""
	Helper function to extract a class from module
	"""
	# Import module
	mod = importlib.import_module(module)
	# Get class
	return getattr(mod, className)

def parserFactory( specs, context, metrics ):
	"""
	Create a parser from the specs dict
	"""

	# Extract class
	cls_name = specs['class']
	del specs['class']

	# Get class
	cls = _class_by_name( cls_name, 'Parser' )

	# Instantiate and configure
	inst = cls( context, metrics )
	inst.configure( specs )

	# Return instance
	return inst

def pipeFactory( specs, context ):
	"""
	Create a pipe from the specs dict
	"""

	# Extract class
	cls_name = specs['class']
	del specs['class']

	# Get class
	cls = _class_by_name( cls_name, 'Pipe' )

	# Instantiate and configure
	inst = cls( context )
	inst.configure( specs )

	# Return instance
	return inst

def aggregateFactory( specs, metric ):
	"""
	Create an aggregator from the specs dict
	"""

	# Extract class
	cls_name = specs['class']
	del specs['class']

	# Get class
	cls = _class_by_name( cls_name, 'Aggregate' )

	# Instantiate and configure
	inst = cls( metric )
	inst.configure( specs )

	# Return instance
	return inst





from robob.metrics import MetricAggregator

class Aggregate(MetricAggregator):
	"""
	Sum aggregator calculates the sum of collected values
	"""

	def configure(self, config):
		"""
		Configure aggregator
		"""
		self.title = "(Sum)"
		if 'title' in config:
			self.title = config['title']

	def collect(self, values):
		"""
		Run aggregator for the specified values and collect results
		"""
		
		# Summarize
		num = 0
		for v in values:
			num += v.number()

		# Return sum
		return [ num ]

	def titles(self):
		"""
		Return the titles of this aggregator values
		"""
		return [ self.title ]




import shlex
from robob.pipe import PipeBase

class Pipe(PipeBase):
	"""
	Implementation of the application pipe item
	"""

	def configure(self, config):
		"""
		Configure cmdline
		"""

		# Prepare
		self.binary = config['binary']
		self.args = []
		self.stdin = None

		# Parse arguments
		if 'args' in config:
			if type(config['args']) in [str, str]:
				self.args = shlex.split(config['args'])
			elif isinstance(config['args'], list):
				self.args = config['args']
			else:
				raise AssertionError("Application's arguments must be a string or list!")

		# Process stdin
		if 'stdin' in config:
			self.stdin = config['stdin']

		# Parse application environment
		self.env = {}
		if 'env' in config:
			n = "env.%s" % config['env']
			if not n in self.context:
				raise AssertionError("Unknown environment '%s' in application specs" % config['env'])
			self.env = self.context[ n ]

	def pipe_stdin(self):
		"""
		Return app stdin buffer (if any)
		"""
		return self.stdin

	def pipe_cmdline(self):
		"""
		Pipe local arguments to command-line
		"""

		# Prepare args
		args = [ self.binary ]
		args += self.args

		# Append child command-lines
		args += super(Pipe, self).pipe_cmdline()

		# If we have environment wrap in 'env'
		if self.env:
			env_prefix = [ 'env' ]
			for k,v in self.env.items():
				env_prefix.append( "%s=%s" % (k,v) )

			# Update with prefix
			args = env_prefix + args

		# Return new arguments
		return args





import logging
import time

from collections import OrderedDict
from robob.factories import aggregateFactory

logger = logging.getLogger("metrics")

#: SI prefix metric (ex. 10^3->k, 10^6->M)
METRIC_SI = 1

#: IEC prefix metric (ex. 1024->k, 1024^2->M)
METRIC_IEC = 2

def _apply_prefix( value, base, prefixes ):
	"""
	Test what's the maximum prefix we can apply to the specified
	value with the specified base
	"""
	nv = value
	np = ""

	# Test given prefixes
	for i in range(1, len(prefixes)):
		v = pow(base, i)

		if v >= 1:
			if v > value:
				break
		else:
			if v < value:
				break

		# Apply scaling & pick prefix
		nv = value / v
		np = prefixes[i]

	# Return tuple
	return (nv, np)

def summarize( results ):
	"""
	Summarize multiple results into a single one
	"""

	# Create a results were to collect everything
	ans = MetricsResults()

	# If no results, return nothing
	if not results:
		return ans

	# Metrics are always the same, so just get the first ones
	ans.metrics = results[0].metrics
	ans.values = [0] * len(results[0].values)

	# Summarize values
	for r in results:
		i = 0
		for a,b in zip(ans.values, r.values):
			if a is None and b is None:
				ans.values[i] = None
			elif a is None:
				ans.values[i] = b+b
			elif b is None:
				ans.values[i] = a+a
			else:
				ans.values[i] = a+b
			i += 1

	# Average
	num = len(results)
	ans.values = [ x / num for x in ans.values ]

	# Return results
	return ans

class MetricAggregator(object):
	"""
	Default metric aggregator
	"""

	def __init__(self, metric):
		"""
		Initialize a metric aggregator
		"""
		self.metric = metric

	def configure(self, config):
		"""
		Configure metrics aggregator
		"""
		pass

	def collect(self, values):
		"""
		Run aggregator over the given timeseries values and
		return an array of values
		"""
		return []

	def titles(self):
		"""
		Return the title suffixes for the values returned by
		the previous function.
		"""
		return []

class MetricValue(object):
	"""
	A value with a timestamp used in the timeseries
	"""

	def __init__(self, value):
		"""
		Keep value and timestamp
		"""
		self.t = time.time()
		self.v = value

	def number(self):
		"""
		Return value as number
		"""
		try:
			if type(self.v) in [str, str]:
				if '.' in self.v:
					return float(self.v)
				else:
					return int(self.v)
			else:
				return self.v
		except ValueError:
			return 0.0

class Metric(object):
	"""
	A single metric on the metrics array
	"""

	def __init__(self, config):
		"""
		Initialize metric
		"""

		# Defaults and required properties
		self.name = config['name']
		self.title = self.name
		self.initial = 0
		self.series = []
		self.units = ""
		self.prefix = 0
		self.scale = 1.0
		self.decimals = 2
		self.aggregators = []
		self.resetTime = 0
		self.unitsInValues = False

		# Update optional
		if 'title' in config:
			self.title = config['title']
		if 'units' in config:
			self.units = config['units']
		if 'showunits' in config:
			self.unitsInValues = config['showunits']
		if 'initial' in config:
			self.initial = config['initial']
		if 'prefix' in config:
			m = config['prefix'].lower()
			if m == 'si':
				self.prefix = METRIC_SI
			elif m == 'iec':
				self.prefix = METRIC_IEC
			else:
				raise AssertionError("Unknown metrix prefix '%s'. Expecting 'si' or 'iec'")
		if 'scale' in config:
			self.scale = float(config['scale'])
		if 'dec' in config:
			self.decimals = int(config['dec'])
		if 'aggregate' in config:
			aggregate = config['aggregate']
			if type(aggregate) in [str, str]:
				aggregate = [ {"class": aggregate} ]
			if type(aggregate) is dict:
				aggregate = [ aggregate ]
			for a in aggregate:
				if type(a) in [str, str]:
					a = {"class": a}
				self.aggregators.append( aggregateFactory(a, self) )

		# If we have no aggregators, add a default
		# average aggregator
		if not self.aggregators:
			self.aggregators.append( aggregateFactory({
					"class": "robob.aggregate.avg"
				}, self) )

		# Reset
		self.reset()

	def update(self, value):
		"""
		Add a value in the time series
		"""
		self.series.append( MetricValue(value) )

	def reset(self):
		"""
		Reset to default
		"""
		self.series = []
		self.resetTime = time.time()

	def format(self, value, withunits=False):
		"""
		Human-readable formatting of the given metric value
		"""

		# Handle none
		if value is None:
			return "(None)"

		# Apply scale to value
		v = value * self.scale
		u = ""

		# If we should show units in the values include them now
		if self.unitsInValues or withunits:

			# Get units
			u = self.units

			# Apply prefix
			if self.prefix == METRIC_SI:
				if v < 1:
					(v, sf) = _apply_prefix( v, 0.001, [ 'm','u','n','p','f','a' ] )
				else:
					(v, sf) = _apply_prefix( v, 1000, [ 'k','M','G','T','P','E' ] )
				u = sf + u
			elif self.prefix == METRIC_IEC:
				(v, sf) = _apply_prefix( v, 1024, [ 'k','M','G','T','P','E' ] )
				u = sf + u

			# Add space
			u = " %s" % u

		# Format decimals
		s = ('{0:.%ig}' % self.decimals).format(v)

		# Return value with units
		return s + u

	def titles(self):
		"""
		Return the titles fo the values returned by this metric
		"""
		titles = []

		# Collect titles from aggregators
		for a in self.aggregators:
			a_titles = a.titles()

			# Skip case were we have 1 aggregator with 1 title
			if (len(a_titles) == 1) and (len(self.aggregators) == 1):
				if not self.unitsInValues and self.units:
					titles.append( "%s [%s]" % (self.title, self.units) )
				else:
					titles.append( self.title)
			else:
				for t in a_titles:
					if not self.unitsInValues and self.units:
						titles.append( "%s %s [%s]" % (self.title, t, self.units) )
					else:
						titles.append( "%s %s" % (self.title, t) )

		# Return titles
		return titles

	def values(self):
		"""
		Return metric values as collected from the aggregators
		"""
		values = []

		# Create values from aggregators
		for a in self.aggregators:
			values += a.collect( self.series )

		# Return values
		return values

class MetricsResults(object):
	"""
	An abstract representation of metrics results that can be summarized
	with other metrics results before finalized to renderable results.
	"""

	def __init__(self):
		"""
		Initialize a results class
		"""

		self.values = []
		self.metrics = []

	def updateFrom(self, metric):
		"""
		Update values of the specified metric
		"""

		# Get aggregated values
		values = list(metric.values())

		# Update value and linked metric for it
		for v in values:
			self.values.append( v )
			self.metrics.append( metric )

	def render(self, withunits=False):
		"""
		Render the results to human-readable indicators
		"""
		results = []

		# Iterate over values and render them
		for i in range(0, len(self.values)):
			results.append( self.metrics[i].format( self.values[i], withunits ) )

		# Return results
		return results

class Metrics(object):
	"""
	Metrics class is responsible for keeping track for metrics
	updates and for summarizing it's results.
	"""

	def __init__(self):
		"""
		Initialize metrics object
		"""
		self.metrics = OrderedDict()

	def configure(self, config):
		"""
		Configure metrics object
		"""

		# Initialize metrics from their config
		for m in config['metric']:
			self.metrics[ m['name'] ] = Metric( m )

	def reset(self):
		"""
		Reset all metrics
		"""
		for m in list(self.metrics.values()):
			m.reset()

	def update(self, name, value):
		"""
		Update the specified value to a metric
		"""

		# Update the specified metric
		if name in self.metrics:
			logger.debug("Updating metric '%s' to '%s'" % (name, str(value)))
			self.metrics[name].update( value )
		else:
			logger.warn("Trying to update an unknown metric: '%s'" % name)

	def titles(self):
		"""
		Return the titles of the metrics
		"""
		titles = []

		# Get the aggregator titles of each metric
		for m in list(self.metrics.values()):
			titles += m.titles()

		# Return titles
		return titles

	def results(self):
		"""
		Collect current results in a result object
		"""

		# Create new metrics results using the specs from the metrics
		results = MetricsResults()

		# Start aggregating results
		for m in list(self.metrics.values()):
			results.updateFrom( m )

		# Return resultset
		return results





import re
from robob.parser import ParserBase

WS = re.compile(r"\s+")
CPULINE = re.compile(r"^([0-9]+):([0-9]+):([0-9]+)")

class Parser(ParserBase):
	"""
	The mpstat parser parses the output of mpstat utility on linux
	"""

	def configure(self, config):
		"""
		Apply the specified configuration
		"""

		# Data matrix
		self.matrix = {}

		# Local properties
		self.core_parsing = False
		self.metrics = []
		self.prefixmatch = ""

		# Extract what information to track
		self.updateMetrics = {}
		if 'match' in config:
			for k,v in config['match'].items():

				# Update metrics
				parts = v.split(".")
				if len(parts) != 2:
					raise AssertionError("Expecting 'cpu.metric' format for the metrics to track")
				self.updateMetrics[k] = parts

	def got_stdout(self, line):
		"""
		Match the specified line against our configuration
		"""

		if "%usr" in line:

			# Prepare for parsing
			self.core_parsing = True
			self.matrix = {}

			# Split header
			header = WS.split(line)
			if header[2] != "CPU":
				raise AssertionError("This does not look like mpstat output!")

			# Extract metric titles
			self.metrics = [ x.replace("%","") for x in header[3:-1] ]

		elif self.core_parsing:

			if CPULINE.match(line):
				parts = WS.split(line)
				
				# Get key and values
				cpu = parts[2]
				values = [ float(x) for x in parts[3:-1] ]

				# Update CPU details on this matrix
				self.matrix[cpu] = dict(list(zip( self.metrics, values )))

			else:
				self.core_parsing = False
				self._commit()


	def got_eof(self):
		"""
		Transmission completed
		"""
		self._commit()

	def _commit(self):
		"""
		Update metrics according to details
		"""

		for metric, (cpu, track) in self.updateMetrics.items():
			value = None

			# Get this value
			if cpu in self.matrix:
				if track in self.matrix[cpu]:
					value = self.matrix[cpu][track]

			# Update metric
			if not value is None:
				self.update( metric, value )





import logging
from robob.pipe import PipeBase, PipeExpect

class Pipe(PipeBase):
	"""
	SSH Tunelling pipe
	"""

	def configure(self, config):
		"""
		Configure access object
		"""

		# Prepare command line
		self.username = config['username']
		self.args = []
		self.password = None
		self.host = None

		# Prepare private key or password
		if 'key' in config:
			self.args += [ '-i', config['key'] ]
		elif 'password' in config:
			self.password = config['password']

		# Override host if defiend
		if 'host' in config:
			self.host = config['host']

	def pipe_cmdline(self):
		"""
		Pipe local arguments to command-line
		"""

		# Get hostname
		host = self.host
		if host is None:
			host = self.context["node.host"]

		# Prepare args
		args = [ "/usr/bin/ssh", "-t", "-q", "-o", "UserKnownHostsFile=/dev/null", "-o", "StrictHostKeyChecking=no" ]
		if self.password:
			args += [ "-o", "PreferredAuthentications=password" ]
		args += [ "%s@%s" % (self.username, host) ]
		args += self.args
		args += [ "--" ]

		# Append child command-lines
		args += super(Pipe, self).pipe_cmdline()

		# Return new arguments
		return args

	def pipe_expect_stdout(self):
		"""
		Add an expect entry to send password when requested
		"""

		# Prepare expect
		expect = []
		if self.password:
			expect.append( PipeExpect( r"[Pp]assword:", send=self.password+"\r" ) )

		# Forward
		return expect + PipeBase.pipe_expect_stdout(self)





import re
from robob.component import ComponentBase

class PipeListener(object):
	"""
	A class that can receives stdout/stderr events
	"""

	def got_stdout(self, line):
		"""
		Process an stdout line
		"""
		pass

	def got_stderr(self, line):
		"""
		Process an stderr line
		"""
		pass

	def got_eof(self):
		"""
		Process end of stream
		"""
		pass

class PipeExpect(object):
	"""
	A pipe expect entry
	"""

	def __init__(self, match, callback=None, send=None):
		"""
		Initialize a pipe expect entry
		"""

		# Initialize properties
		self.repr = match
		self.match = re.compile(match)
		self.callback = callback
		self.send = send

		# Match function result
		self.found = None

	def matches(self, line):
		"""
		Test if the pipe expects
		"""

		# Return the match if found, None otherwise
		self.found = self.match.search( line )
		return self.found

	def render(self):
		"""
		Render the result to send after expect
		"""

		# Return data if only data specified
		if self.send:
			return self.send

		# Otherwise use callback
		elif self.callback:
			return self.callback( self.found )

		# Nothing found? Return None
		else:
			return ""

	def __str__(self):
		return self.repr

class PipeBase(ComponentBase):
	"""
	A chainable pipe object
	"""

	def __init__(self, context):
		"""
		Initialize pipe base
		"""
		super(PipeBase, self).__init__(context)

		self.pipes = []
		self.listeners = []

	def configure(self, specs):
		"""
		Configure specs
		"""
		pass

	def pipe_stdin(self):
		"""
		Return the pipe's stdin buffer
		"""
		stdin = ""
		for p in self.pipes:
			stdin += p.pipe_stdin()
		return stdin

	def pipe_cmdline(self):
		"""
		Return pipe's cmdline
		"""
		cmdline = []
		for p in self.pipes:
			cmdline += p.pipe_cmdline()
		return cmdline

	def pipe_stdout(self, stdout):
		"""
		Handle stdout line
		"""

		# Trigger to listeners
		for l in self.listeners:
			l.got_stdout( stdout )

		# Forward to children
		for p in self.pipes:
			p.pipe_stdout( stdout )

	def pipe_stderr(self, stderr):
		"""
		Handle stderr line
		"""

		# Trigger to listeners
		for l in self.listeners:
			l.got_stderr( stderr )

		# Forward to children
		for p in self.pipes:
			p.pipe_stderr( stderr )

	def pipe_expect_stdout(self):
		"""
		Collect expect entries on stdout
		"""
		entries = []
		for p in self.pipes:
			entries += p.pipe_expect_stdout()
		return entries

	def pipe_expect_stderr(self):
		"""
		Collect expect entries on stdout
		"""
		entries = []
		for p in self.pipes:
			entries += p.pipe_expect_stderr()
		return entries

	def pipe_close(self):
		"""
		Request to close all pipe resources
		"""

		# Trigger to listeners
		for l in self.listeners:
			l.got_eof()

		# Forward to children
		for p in self.pipes:
			p.pipe_close()

	def plug(self, pipe):
		"""
		Plug a pipe in the this
		"""
		if not isinstance(pipe, PipeBase):
			raise AssertionError("The given pipe object is not instance of PipeBase")
		self.pipes.append(pipe)

	def listen(self, listener):
		"""
		Listen for input events
		"""
		if not isinstance(listener, PipeListener):
			raise AssertionError("The given listener object is not instance of PipeListener")
		self.listeners.append(listener)




import datetime
import logging
from collections import OrderedDict

class Reporter(object):
	"""
	A class tha generates reports
	"""

	def __init__(self, filename, specs):
		"""
		Initialize a new test report, keeping the data in the output specified
		"""

		self.fd = None
		self.testID = 0
		self.specs = specs
		self.filename = filename
		self.iterations = int(specs.specs.get("test.iterations", 1))
		self.testVariables = specs.getTestVariables()
		self.testTitles = specs.getMetricTitles()
		self.activeTest = []
		self.summaryLines = []

		self.in_iteration = False
		self.in_test = False
		self.cur_iterations = 0
		self.ok_iterations = 0

		# Calculate maximum title width
		self.testTitleWidth = 1
		for t in self.testTitles:
			if len(t) > self.testTitleWidth:
				self.testTitleWidth = len(t)

		# Open logger
		self.logger = logging.getLogger("report")

		# Collect metadata
		self.notes = OrderedDict()
		if 'title' in specs.specs:
			self.notes['Title'] = specs.specs['title']
		if 'desc' in specs.specs:
			self.notes['Description'] = specs.specs['desc']
		if 'notes' in specs.context:
			self.notes.update( specs.context['notes'] )

	def start( self ):
		"""
		Start a test putting metadata and columns in the file
		"""

		# Reset properties
		self.testID = 0
		self.activeTest = []

		# Open file descriptor
		self.logger.info("Writing report to %s" % self.filename)
		self.fd = open(self.filename, "w")

		# Write title & Columns
		for k,v in self.notes.items():
			self.fd.write("%s,%s\n" % (k, v))
		self.fd.write("Started on,%s\n" % str(datetime.datetime.now()))
		self.fd.write("\n")
		self.fd.write("Test numbers\n")
		self.fd.write("\n")
		self.fd.write("Num,Iteration,Started,Ended,Status,%s,%s,Comment\n" % \
			( ",".join(self.testVariables),  ",".join(self.testTitles) ) )

	def close(self):
		"""
		Close test report
		"""

		# Close file descriptor if open
		if self.fd:
			self.fd.close()
			self.fd = None

	def finalize(self):
		"""
		Finalize report
		"""

		self.logger.info("Finalizing report")

		self.fd.write("\n")
		self.fd.write("Summarized numbers\n")
		self.fd.write("\n")
		self.fd.write("Num,Started,Ended,Iterations,Successful,%s,%s,Comment\n" % \
			( ",".join(self.testVariables),  ",".join(self.testTitles) ) )

		# Write summarization lines
		for l in self.summaryLines:
			self.fd.write(l)

		# Flush
		self.fd.flush()

	def iteration_start( self, iteration ):
		"""
		Log the start of a test
		"""

		# Log the beginning of test and starting date
		self.fd.write("%i,%i of %i,%s" % \
			( self.testID, iteration, self.iterations, str(datetime.datetime.now()) ) )
		self.fd.flush()

		# Enter iteration
		self.in_iteration = True
		self.cur_iterations += 1

	def iteration_end(self, results, status="Completed", comment="" ):
		"""
		Log the completion of a test
		"""

		# Write end and values
		self.fd.write(",%s,%s,%s,%s,%s\n" % \
			( str(datetime.datetime.now()), status, ",".join(self.activeTest), ",".join(results.render()), comment ) )
		self.fd.flush()

		# Count successful iterations
		if status == "Completed":
			self.ok_iterations += 1

		# Exit iteration
		self.in_iteration = False

		# Print values
		rendered = results.render( True )
		self.logger.info( "-" * (self.testTitleWidth + 20) )
		for i in range(0, len(self.testTitles)):
			self.logger.info(
				(("%%%is : ") % self.testTitleWidth) % self.testTitles[i] + rendered[i]
			)
		self.logger.info( "-" * (self.testTitleWidth + 20) )

	def test_start( self, testContext ):
		"""
		Log the start of a groupped test
		"""

		# Reset iterations
		self.iterations = int(testContext.get("test.iterations", 1))
		self.cur_iterations = 0
		self.ok_iterations = 0

		# Prepare properties
		self.testID += 1
		self.activeTest = [ str(testContext[x]) for x in self.testVariables ]
		self.in_test = True

		# Keep for summary
		self.summaryLines.append(
				"%i,%s" % ( self.testID, str(datetime.datetime.now()) )
			)

	def test_end( self, results, comment="" ):
		"""
		Log the end of a groupped test
		"""

		# Write end and values
		self.summaryLines[ len(self.summaryLines)-1 ] += \
			",%s,%i,%i,%s,%s,%s\n" % ( str(datetime.datetime.now()), self.cur_iterations, self.ok_iterations, ",".join(self.activeTest), ",".join(results.render()), comment ) 

		self.in_test = False


	def interrupt(self, results, reason="Interrupted by the user"):
		"""
		User interrupted the test, log the action
		"""

		# Finalize iterations
		if self.in_iteration:
			self.fd.write(",%s,%s,%s,%s,%s\n" % \
				( str(datetime.datetime.now()), "Interrupted", ",".join(self.activeTest), ",".join( [""] * len(self.testTitles) ), reason ) )

		# Finalize test
		if self.in_test:
			if not results:
				self.fd.write(",%s,%s,%s,%s,%s\n" % \
					( str(datetime.datetime.now()), "Interrupted", ",".join(self.activeTest), ",".join( [""] * len(self.testTitles) ), reason ) )
			else:
				self.summaryLines[ len(self.summaryLines)-1 ] += \
					",%s,%i,%i,%s,%s,%s\n" % ( str(datetime.datetime.now()), self.cur_iterations, self.ok_iterations, ",".join(self.activeTest), ",".join(results.render()), reason ) 

		# Finalize
		self.finalize()





import re
from robob.parser import ParserBase

class Parser(ParserBase):
	"""
	The regular expression parser matches each line against
	a set of regex rules and extracts the matches into metrics.

	Example:

	  class: robob.parser.regex
	  match:
	    - "^.*?Got (?P<CQ>[0-9]+) CQ.*$"

	"""

	def configure(self, config):
		"""
		Apply the specified configuration
		"""
		self.matches = []

		# Separate specified configuration to per-line
		for cfg in config['match']:
			self.matches.append( re.compile(cfg) )

	def got_stdout(self, line):
		"""
		Match the specified line against our configuration
		"""

		for m in self.matches:

			# Check for match
			grp = m.match(line)
			if grp:

				# Update values
				for k,v in grp.groupdict().items():
					self.update( k, float(v) )





import re
from robob.parser import ParserBase

class Parser(ParserBase):
	"""
	The simple split parser gives you the ability to extract a particular
	column from a specific line in the output.

	  class: robob.parser.grid 
	  separator: \s+				# How to split each line
	  anchor: 						# Keyword were line index resets to 0
	  match:
	    - name: sample1				# Match on exact line/column
	      line: 1
	      col: 1

	    - name: sample2				# Match column on every line
	      col: 1

	    - name: sample3				# Get the column of a line that matches
	      anchor: "name:"			# the given anchor
	      col: 1

	"""

	def configure(self, config):
		"""
		Apply the specified configuration
		"""
		self.line = 0
		self.col_any = []
		self.col_line = {}
		self.col_anchor = []

		# Column separator
		self.colsep = re.compile(r"\s+")
		if 'separator' in config:
			self.colsep = re.compile(config['separator'])


		# anchor indicator
		self.anchor = None
		if 'anchor' in config:
			self.anchor = re.compile(config['anchor'])

		# Separate specified configuration to per-line
		for cfg in config['match']:

			# Get column separator
			colsep = self.colsep
			if 'separator' in cfg:
				colsep = re.compile(cfg['separator'])

			# Check per-line config
			if 'line' in cfg:
			
				# Ensure a col_line entry
				l = str(cfg['line'])
				if not l in self.col_line:
					self.col_line[l] = []

				# Append
				self.col_line[l].append( (colsep, int(cfg['col']), cfg['name']) ) 

			# Check per-anchor config
			elif 'anchor' in cfg:

				# Append to anchor config
				self.col_anchor.append( ( re.compile(cfg['anchor']), colsep, int(cfg['col']), cfg['name']) )

			# No line specification? Match all
			else:

				# Append to any line config
				self.col_any.append( (colsep, int(cfg['col']), cfg['name']) )


	def got_stdout(self, line):
		"""
		Match the specified line against our configuration
		"""

		# Check for line anchor indicator
		if self.anchor and self.anchor.match(line):
			self.line = 0

		# Process anyline config
		for sep, col, name in self.col_any:

			# Get value
			parts = sep.split(line)
			if col < len(parts):
				self.update( name, parts[col] )

		# Get line as string
		l = str(self.line)
		self.line += 1

		# Check if we have a configuration for this line
		if l in self.col_line:
			for sep, col, name in self.col_line[l]:

				# Get value
				parts = sep.split(line)
				if col < len(parts):
					self.update( name, parts[col] )

		# Process lines that contain a known anchor
		for anchor, sep, col, name in self.col_anchor:
			if anchor.match( line ):

				# Get value
				parts = sep.split(line)
				if col < len(parts):
					self.update( name, parts[col] )


	def reset(self):
		"""
		When reset, anchor lines to line #0
		"""
		self.line = 0




"""A setuptools based setup module.
See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='robob',
    version='0.1.0b1',

    description='A python tool for simplifying collection of measurements over repetitive tasks',
    long_description=long_description,

    # The project's main homepage.
    url='https://github.com/wavesoft/robob',

    # Author details
    author='Ioannis Charalampidis',
    author_email='johnys2@gmail.com',

    # License
    license='Apache-2.0',

    # Classifiers
    classifiers=[

        # Project maturity
        'Development Status :: 4 - Beta',

        # Indicate who your project is intended for
        'Intended Audience :: Science/Research',
        'Intended Audience :: System Administrators',

        # Topics
        'Topic :: Software Development :: Quality Assurance',
        'Topic :: Software Development :: Testing',
        'Topic :: Utilities',

        # License
        'License :: OSI Approved :: Apache Software License',

        # Python versions (tested)
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],

    # What does your project relate to?
    keywords='automation benchmark scripting test',

    # Packages
    packages=find_packages(exclude=['contrib', 'docs', 'tests']),

    # Project dependencies
    install_requires=['PyYAML'],

    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    entry_points={
        'console_scripts': [
            'robob=robob.cli:main',
        ],
    },
)




from robob.metrics import MetricAggregator

class Aggregate(MetricAggregator):
	"""
	Average aggregator calculates the minimum of the collected values
	"""

	def configure(self, config):
		"""
		Configure aggregator
		"""
		self.title = "(Min)"
		if 'title' in config:
			self.title = config['title']

	def collect(self, values):
		"""
		Run aggregator for the specified values and collect results
		"""

		# Summarize
		num = None
		if len(values) > 0:
			num = values[0].v
			for v in values:
				n = v.number()
				if n < num:
					num = n

		# Return minimum
		return [ num ]

	def titles(self):
		"""
		Return the titles of this aggregator values
		"""
		return [ self.title ]




from robob.metrics import MetricAggregator

#: The metric shows how many bytes are transfered till now
#: (ex. 10, 20, 50, 100, etc.)
MODE_INCREMENTING = 0

#: The metric shows how many bytes are transfered per operation
#: (ex. 10, 10, 30, 50, etc..)
MODE_PARTIAL = 1

#: The metric shows how many transfer operations are completed
#: (ex. 1, 1, 3, 5, etc...)
MODE_OPERATIONS = 2

class Aggregate(MetricAggregator):
	"""
	Calculate bandwidth based on how many bytes are transfered over time
	"""

	def __init__(self, metric):
		"""
		Initialize aggregator
		"""
		MetricAggregator.__init__(self, metric)
		self.mode = MODE_INCREMENTING
		self.opsize = 1

	def configure(self, config):
		"""
		Configure bandwidth calculator
		"""

		MODES = {
			"incrementing": MODE_INCREMENTING,
			"partial": MODE_PARTIAL,
			"operations": MODE_OPERATIONS,
			"0": MODE_INCREMENTING,
			"1": MODE_PARTIAL,
			"2": MODE_OPERATIONS,
		}

		if 'mode' in config:
			m = config['mode'].lower()
			if not m in MODES:
				raise AssertionError("Unknown operation mode '%s'" % m)
			self.mode = MODES[ m ]

		if 'opsize' in config:
			self.opsize = float(config['opsize'])

	def collect(self, values):
		"""
		Run aggregator for the specified values and collect results
		"""

		# Min max and avereage
		v_min = None
		v_max = None
		v_avg = 0

		# Get lat value
		if self.mode == MODE_PARTIAL:
			last_v = self.metric.initial
			last_t = self.metric.resetTime
		else:
			last_v = None
			last_t = None

		# Iterate over values
		for v in values:
			if last_v:

				# Calculate time difference (sec)
				time_diff = v.t - prev.t

				# Calculate bandwidth
				if self.mode == MODE_INCREMENTING:
					bw_diff = last_v - v.v
				elif self.mode == MODE_PARTIAL:
					bw_diff = v.v
				elif self.mode == MODE_OPERATIONS:
					bw_diff = v.v * self.opsize

				# Calculate bandwidth (bytes/sec)
				bw = bw_diff / time_diff

				# Update
				if v_min is None:
					v_min = bw
					v_max = bw
					v_avg = bw
				else:
					if bw < v_min:
						v_min = bw
					if bw > v_max:
						v_max = bw
					v_avg += bw

			last_v = v.v
			last_t = v.t

		# Average
		v_avg /= len(values)

		# Return bandwidth suffixes
		return [ v_avg, v_min, v_max ]

	def titles(self):
		"""
		Return the titles of this aggregator values
		"""
		return [ "(Average B/w)", "(Min B/w)", "(Max B/w)" ]





import sys
import logging
import random
import string
import shlex
import subprocess
from robob.pipe import PipeBase

def list2cmdline(seq):
	"""
	Modified version of the original from subprocess.py
	"""
	result = []
	needquote = False
	for arg in seq:
		bs_buf = []

		# Add a space to separate this argument from the others
		if result:
			result.append(' ')

		needquote = (" " in arg) or ("\t" in arg) or (";" in arg) or not arg
		if needquote:
			result.append('"')

		for c in arg:
			if c == '\\':
				# Don't know if we need to double yet.
				bs_buf.append(c)
			elif c == '"':
				# Double backslashes.
				result.append('\\' * len(bs_buf)*2)
				bs_buf = []
				result.append('\\"')
			else:
				# Normal char
				if bs_buf:
					result.extend(bs_buf)
					bs_buf = []
				result.append(c)

		# Add remaining backslashes, if any.
		if bs_buf:
			result.extend(bs_buf)

		if needquote:
			result.extend(bs_buf)
			result.append('"')

	return ''.join(result)

class Pipe(PipeBase):
	"""
	Implementation of the bash wrapper that takes 
	"""

	def __init__(self, ctx):
		"""
		Initialize pipe
		"""
		super(Pipe, self).__init__(ctx)

		# Init properties
		self.script_blocks = []
		self.pipes_pre = []
		self.pipes_post = []
		self.logger = logging.getLogger("pipe.bashwrap")

	def plug_pre(self, pipe):
		"""
		Plug a pipe in the beginning of the script
		"""
		self.pipes_pre.append( pipe )

	def plug_post(self, pipe):
		"""
		Plug a pipe in the end of the script
		"""
		self.pipes_post.append( pipe )

	def pipe_cmdline(self):
		"""
		Return piped command-line
		"""

		# We are piping everything to bash
		return [ "/usr/bin/stdbuf", "-oL", "-eL", "/bin/bash", "/dev/stdin" ]

	def pipe_stdin(self):
		"""
		Return piped stdin
		"""

		# Definitions
		s_defs = ""
		s_run = ""
		s_pre = ""
		s_post = ""
		s_killtrap = "function killer_@@ {\n"

		# Render pre-pipes
		for p in self.pipes_pre:

			# Get fragment script
			cmdline = p.pipe_cmdline()
			if cmdline[0] == "eval":
				frag_script = " ".join(cmdline[1:])
			else:
				frag_script = list2cmdline(cmdline)

			# Add semi-colon at the end of the fragment if we don't already
			# have a command terminator
			if not frag_script[-1] in ["\n", ";"]:
				frag_script += ";"

			# Include fragment in preconditions
			s_pre += "{ %s } >/dev/null 2>/dev/null\n" % frag_script
			s_pre += "[ $? -ne 0 ] && echo \"::W::A pre-condition failed\"\n"

		# Render post-pipes
		for p in self.pipes_post:

			# Get fragment script
			cmdline = p.pipe_cmdline()
			if cmdline[0] == "eval":
				frag_script = " ".join(cmdline[1:])
			else:
				frag_script = list2cmdline(cmdline)

			# Add semi-colon at the end of the fragment if we don't already
			# have a command terminator
			if not frag_script[-1] in ["\n", ";"]:
				frag_script += ";"

			# Include fragment in postconditions
			s_post += "{ %s } >/dev/null 2>/dev/null\n" % frag_script
			s_post += "[ $? -ne 0 ] && echo \"::W::A post-condition failed\"\n"

		# Prepare pipe chunks
		for i in range(0, len(self.pipes)):
			p = self.pipes[i]

			# Prepare prefix for this pipe
			prefix = "::%i::" % i

			# Get fragment script
			cmdline = p.pipe_cmdline()
			if cmdline[0] == "eval":
				frag_script = " ".join(cmdline[1:])
			else:
				frag_script = list2cmdline(cmdline)

			# Add semi-colon at the end of the fragment if we don't already
			# have a command terminator
			if not frag_script[-1] in ["\n", ";"]:
				frag_script += ";"

			# Define fragment with prefixed stdout & stderr
			s_defs += "function frag_%i {\n" % (i,)
			s_defs += "{ { %s } 2>&3 | awk >&2 '$0=\"%s\"$0'; exit ${PIPESTATUS[0]}; } 3>&1 1>&2 | awk '$0=\"%s\"$0';\n" % \
				(frag_script, prefix, prefix)
			s_defs += "return ${PIPESTATUS[0]}\n"
			s_defs += "}\n"

			# Define runner function
			s_defs += "function run_%i {\n" % (i,)
			inbuf = p.pipe_stdin()
			if inbuf:
				eof_indicator = "STDIN%i_" % (i,)
				eof_indicator += ''.join(random.choice(string.ascii_uppercase + string.digits) for _ in range(16))
				s_defs += "cat <<'%s' | frag_%i\n" % (eof_indicator,i)
				s_defs += inbuf
				s_defs += "\n%s" % eof_indicator
			else:
				s_defs += "frag_%i;" % (i,)
			s_defs += "\n}\n"

			# Define run script
			s_run += "run_%i&\nFRAG_PID_%i=$!\n" % (i,i)

			# Define killer trap
			s_killtrap += "kill -@@ $FRAG_PID_%i 2>/dev/null\n" % (i,)

		# Finalize fragments
		s_killtrap += "trap - @@\n"
		s_killtrap += "}\ntrap killer_@@ @@\n"

		# Compile script
		script = "# Definitions\n"
		script += s_defs
		script += "# Signal hooks\n"
		script += s_killtrap.replace("@@", "SIGINT")
		script += s_killtrap.replace("@@", "SIGHUP")
		script += s_killtrap.replace("@@", "SIGKILL")
		if s_pre:
			script += "# Pre-conditions\n"
			script += "echo ::D::Satisfying pre-conditions\n"
			script += s_pre
		if s_post:
			script += "# Post-conditions\n"
			script += "function exit_handler {\n"
			script += "echo ::D::Satisfying post-conditions\n"
			script += s_post
			script += "}\n"
			script += "trap exit_handler EXIT\n"
		script += "# Run script\n"
		script += "echo ::I::Starting application\n"
		script += s_run
		script += "# Wait for first fragment complete\n"
		script += "wait $FRAG_PID_0\n"
		script += "RET=$?\n"
		script += "echo ::I::Application exited with code=$RET\n"
		script += "# Interrupt the rest\n"
		script += "killer_SIGINT\n"
		script += "exit $RET\n"

		# sys.stdout.write("----BEGIN SCRIPT----\n%s\n----END SCRIPT----" % script)

		# Return script
		return script

	def pipe_stdout(self, stdout):
		"""
		Forward stdout line to the appropriate pipe
		"""

		# Trigger to local listeners
		self.logger.debug("Handling '%s'" % stdout)
		for l in self.listeners:
			l.got_stdout( stdout )

		# Validate line
		if stdout[0:2] != "::":
			self.logger.debug("Ignoring line (Missing prefix): %s" % stdout)
			return

		# Find end
		end = stdout.find("::",2)
		if end < 0:
			self.logger.debug("Ignoring line (Missing suffix): %s" % stdout)
			return

		# Extract line code
		lc = stdout[2:end]
		if lc == "D":
			self.logger.debug(stdout[end+2:])
			return
		if lc == "I":
			self.logger.info(stdout[end+2:])
			return
		elif lc == "W":
			self.logger.warn(stdout[end+2:])
			return
		elif lc == "E":
			self.logger.error(stdout[end+2:])
			return

		# Extract ID
		uid = int(lc)
		if uid >= len(self.pipes):
			self.logger.debug("Ignoring line (Invalid pipe ID): %s" % stdout)
			return

		# Forward to the correct pipe
		self.logger.debug("Forwarding '%s' to pipe #%i" % ( stdout[end+2:], uid ))
		self.pipes[uid].pipe_stdout( stdout[end+2:] )

